'''
The following was modified by r0s1ta fu.
03-01-2020
github.com/atisor73
                 .__....._                  _.....__,
                      .": o :':          ;': o :".
                      `. `-' .'.        .'. `-' .'
                        `---'              `---'
             _...----...       ...   ...       ...----..._
          .-'__..-""'----     `. `ρζ`  .'     ----'""-..__`-.
         '.-'   _.--"""'        `-._.-'        '"""--._   `-.`
         '  .-"'                   :                   `"-.  `
           '   `.               _.'"'._               .'   `
                 `.        ,.-'"       "'-.,        .'
                   `.                            .'
                     `-._                   _.-'
                         `"'--...___...--'"`'

'''

'''
    The MIT License (MIT)

    Copyright (c) 2013-2014 California Institute of Technology

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in
    the Software without restriction, including without limitation the rights to
    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
    the Software, and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
    FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
    COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
    IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

__author__ = "Angela Gong (anjoola@anjoola.com)"

USAGE = '''
===========
   USAGE
===========

>>> import sim
>>> sim.run([graph], [dict with keys as names and values as a list of nodes])

Returns a dictionary containing the names and the number of nodes they got.

Example:
>>> graph = {"2": ["6", "3", "7", "2"], "3": ["2", "7, "12"], ... }
>>> nodes = {"strategy1": ["1", "5"], "strategy2": ["5", "23"], ... }
>>> sim.run(graph, nodes)
>>> {"strategy1": 243, "strategy6": 121, "strategy2": 13}

Possible Errors:
- KeyError: Will occur if any seed nodes are invalid (i.e. do not exist on the
            graph).
'''

from collections import Counter, OrderedDict
from copy import deepcopy
from random import randint

def run(adj_list, node_mappings):
  """
  Function: run
  -------------
  Runs the simulation on a graph with the given node mappings.

  adj_list: A dictionary representation of the graph adjacencies.
  node_mappings: A dictionary where the key is a name and the value is a list
                 of seed nodes associated with that name.
  """
  results, history = run_simulation(adj_list, node_mappings)
  return results, history

def run_simulation(adj_list, node_mappings):
  """
  Function: run_simulation
  ------------------------
  Runs the simulation. Returns a dictionary with the key as the "color"/name,
  and the value as the number of nodes that "color"/name got.

  adj_list: A dictionary representation of the graph adjacencies.
  node_mappings: A dictionary where the key is a name and the value is a list
                 of seed nodes associated with that name.
  """
  node_colors = []
  # Stores a mapping of nodes to their color.
  node_color = dict([(node, None) for node in adj_list.keys()])
  init(node_mappings, node_color)

  generation = 1
  # Keep calculating the epidemic until it stops changing. Randomly choose
  # number between 100 and 200 as the stopping point if the epidemic does not
  # converge.
  prev = None
  nodes = adj_list.keys()
  last_iter = randint(100, 200)
  while not is_stable(generation, last_iter, prev, node_color):
    prev = deepcopy(node_color)
    for node in nodes:
      (changed, color) = update(adj_list, prev, node)
      # Store the node's new color only if it changed.
      if changed:
          node_color[node] = color

    node_colors.append(prev)
    # NOTE: prev contains the state of the graph of the previous generation,
    # node_colros contains the state of the graph at the current generation.
    # You could check these two dicts if you want to see the intermediate steps
    # of the epidemic.
    generation += 1

  return get_result(node_mappings.keys(), node_color), node_colors


def init(color_nodes, node_color):
  """
  Function: init
  --------------
  Initializes the node to color mappings.
  """
  for (color, nodes) in color_nodes.items():
    for node in nodes:
      if node_color[node] is not None:
        node_color[node] = "__CONFLICT__"
      else:
        node_color[node] = color
  for (node, color) in node_color.items():
    if color == "__CONFLICT__":
      node_color[node] = None


def update(adj_list, node_color, node):
  """
  Function: update
  ----------------
  Updates each node based on its neighbors.
  """
  neighbors = adj_list[node]
  colored_neighbors = list(filter(None, [node_color[x] for x in neighbors]))
  team_count = Counter(colored_neighbors)
  if node_color[node] is not None:
    team_count[node_color[node]] += 1.5
  most_common = team_count.most_common(1)
  if len(most_common) > 0 and \
    most_common[0][1] > (len(colored_neighbors) + (1.5 if node_color[node] is not None else 0)) / 2.0:
    return (True, most_common[0][0])

  return (False, node_color[node])


def is_stable(generation, max_rounds, prev, curr):
  """
  Function: is_stable
  -------------------
  Checks whether or not the epidemic has stabilized.
  """
  if generation <= 1 or prev is None:
    return False
  if generation == max_rounds:
    return True
  for node, color in curr.items():
    if not prev[node] == curr[node]:
      return False
  return True


def get_result(colors, node_color):
  """
  Function: get_result
  --------------------
  Get the resulting mapping of colors to the number of nodes of that color.
  """
  color_nodes = {}
  for color in colors:
    color_nodes[color] = 0
  for node, color in node_color.items():
    if color is not None:
      color_nodes[color] += 1
  return color_nodes


if __name__ == '__main__':
  print(USAGE)



import numpy as np
import pandas as pd

try:
    import iqplot
except:
    pass

import bokeh
import bokeh.io
import panel as pn
pn.extension()

import json
import networkx as nx
from . import sim_viz



def style(p):
    p.title.text_font="Helvetica"
    p.title.text_font_size="16px"
    p.title.align="center"
    p.xaxis.axis_label_text_font="Helvetica"
    p.yaxis.axis_label_text_font="Helvetica"

    p.xaxis.axis_label_text_font_size="13px"
    p.yaxis.axis_label_text_font_size="13px"
    p.xaxis.axis_label_text_font_style = "normal"
    p.yaxis.axis_label_text_font_style = "normal"
    p.background_fill_alpha = 0

    p.xaxis.visible,p.yaxis.visible = False, False
    p.xgrid.grid_line_color,p.ygrid.grid_line_color = None, None
    return p

def simulate(graph, l_seed):
    """ Runs Simulation from sim.py
        --------------------------------------------
        graph : either dictionary or nx Graph object
        l_seed : list of seed nodes
    """
    if type(graph) == nx.classes.graph.Graph:
        graph = nx.to_dict_of_lists(graph)  # change graph object to dictionary

    d_seed = dict(zip([f"{i}" for i, _ in enumerate(l_seed)], l_seed))

    return sim_viz.run(graph, d_seed)

def read_opponents(file_path):
    with open(file_path, 'r') as j:
        _opponents = json.loads(j.read())
    opp_names = list(_opponents.keys())
#     opp_names.remove('Nooodles')
    _ = [_opponents[opp] for opp in opp_names]
    opponents = dict(zip(opp_names, _))
    return opponents

def dataframer(positions):
    d = {'nodes':[], 'x':[], 'y':[]}
    for k, v in positions.items():
        d['nodes'].append(k)
        d['x'].append(v[0])
        d['y'].append(v[1])
    df = pd.DataFrame(d)
    return df


def viz(G, seeds,
        palette=['#cb4f70', '#1b718c', '#779e1a', '#f49044',
                '#46308d', '#2ea58e', '#e97d86', '#7e92bd',
                '#c21019', '#f56327', '#fab01d', '#026f69',
                '#750a2b', '#1b9e77', '#d95f02', '#7570b3',
                '#e7298a', '#66a61e'],
        uncolored_size=2.5,
        colored_size=4.5,
        names=None,
        layout=None,
        ):
    """
    Arguments
    -------------------------------------
    G : networkx Graph object
    seeds : list of list of seed nodes to play on G
    palette : list of strings (hex values or html names)
    uncolored_size : float, size of uncolored glyphs
    colored_size : float, size of colored glyphs
    names : list of opponent names for legend labels
    layout : string ("radial", "random", "kamada", "spectral"),
             default "spring"

    Returns :
    -------------------------------------
    Bokeh panel object displaying color cascade, with iteration and zoom control.
    """
    if layout == "radial":     layout = nx.circular_layout(G)
    elif layout == "random":   layout = nx.random_layout(G)
    elif layout == "kamada":   layout = nx.kamada_kawai_layout(G)
    elif layout == "spectral": layout = nx.spectral_layout(G)
    else:                      layout = nx.spring_layout(G)

    df = dataframer(layout)
    result, history = simulate(G, seeds)

    color_map = {k:v for k, v in zip(result.keys(), palette)}
    color_map[None] = "grey"
    print()
    size_map = {None : uncolored_size}
    for k in result.keys():
        size_map[k] = colored_size

    i_slider = pn.widgets.IntSlider(start=1, end=len(history), value=1,
                                    name="iteration", width=370)
    range_slider = pn.widgets.FloatSlider(name='zoom', width=370,
                                      start=0.10, end=2.0, value=1.1, step=0.05)
    @pn.depends(i_slider.param.value, range_slider.param.value)
    def plotter(i, z):
        z = 2.05 - z
        df_ = df
        p = bokeh.plotting.figure(title=f"iteration {i}",
                                  width=600, height=450,
                                  x_range=[-z,z],y_range=[-z, z],  )

        # iterating, update color and size of node
        labels = list(history[i-1].values())
        df_["color"] = [color_map[label] for label in labels]
        df_["size"] = [size_map[label] for label in labels]
        df_ = df_.sort_values(by=["size"])

        # creating legend
        if names==None:
            items = [(f"team {k}", [p.circle(0,0, alpha=0, color=f"{v}")])
                                        for k, v in color_map.items()]
        else:
            items = [(f"{list(names)[i]}", [p.circle(0,0, alpha=0, color=f"{kv[1]}")])
                                        for i, kv in enumerate(color_map.items())
                                        if i < len(seeds)
                                        ]
        legend = bokeh.models.Legend(items=items, location="center")
        p.add_layout(legend, 'right')

        # only plotting nodes
        p.circle(source=df_, x='x',y='y', color='color', size='size',
                 line_color="white", line_width=0.1)

        return style(p)
    return pn.Column(range_slider, i_slider, plotter)


def viz_tournament(G, players,
        palette=['#cb4f70', '#1b718c', '#779e1a', '#f49044',
                '#46308d', '#2ea58e', '#e97d86', '#7e92bd',
                '#c21019', '#f56327', '#fab01d', '#026f69',
                '#750a2b', '#1b9e77', '#d95f02', '#7570b3',
                '#e7298a', '#66a61e'],
        uncolored_size=2.5,
        colored_size=4.5,
        layout=None,
        ):
    """
    Arguments
    -------------------------------------
    G : networkx Graph object
    players : dictionary of {player names: chosen seeds}
    palette : list of strings (hex values or html names)
    uncolored_size : float, size of uncolored glyphs
    colored_size : float, size of colored glyphs
    layout : string ("radial", "random", "kamada", "spectral"),
             default "spring"

    Returns :
    -------------------------------------
    Bokeh panel object displaying color cascade, with cycle+iteration, and zoom control.
    """
    # ***************** GRAPH LAYOUT *****************
    print("graph layout... ")
    if layout == "radial":     layout = nx.circular_layout(G)
    elif layout == "random":   layout = nx.random_layout(G)
    elif layout == "kamada":   layout = nx.kamada_kawai_layout(G)
    elif layout == "spectral": layout = nx.spectral_layout(G)
    else:                      layout = nx.spring_layout(G)

    df = dataframer(layout)

    # ********* SIMULATING 50 CYCLES LAYOUT *********
    print("simulating... ")
    names = list(players.keys())
    all_seeds = list(players.values())

    N_CYCLES = 50
    results, histories = [], []
    for cycle in range(N_CYCLES):
        seeds = [team[cycle] for team in all_seeds]
        result, history = simulate(G, seeds)
        results.append(result)
        histories.append(history)


    # ****************** PLOTTING ******************
    print("plotting... ")
    color_map = {k:v for k, v in zip(result.keys(), palette)}
    color_map[None] = "grey"
    print()
    size_map = {None : uncolored_size}
    for k in result.keys():
        size_map[k] = colored_size

    cycle_selector = pn.widgets.Select(name="cycle", width=90,
        options=list(np.arange(N_CYCLES)), value=0)
    i_slider = pn.widgets.IntSlider(start=1, end=len(histories[0]), value=1,
                                    name="iteration", width=290)
    range_slider = pn.widgets.FloatSlider(name='zoom', width=400,
                                      start=0.10, end=2.0, value=1.1, step=0.05)

    @pn.depends(cycle_selector.param.value, watch=True)
    def update_i_end(cycle):
        i_slider.end = len(histories[cycle])

    @pn.depends(cycle_selector.param.value, i_slider.param.value, range_slider.param.value)
    def plotter(cycle, i, z):
        result, history = results[cycle], histories[cycle]
        z = 2.05 - z
        df_ = df
        p = bokeh.plotting.figure(title=f"iteration {i}",
                                  width=600, height=450,
                                  x_range=[-z,z],y_range=[-z, z],  )

        max_i = len(history)-1
        _index = max_i if i > max_i else i-1

        # iterating, update color and size of node
        labels = list(history[_index].values())
        df_["color"] = [color_map[label] for label in labels]
        df_["size"] = [size_map[label] for label in labels]
        df_ = df_.sort_values(by=["size"])

        # creating legend
        if names==None:
            items = [(f"team {k}", [p.circle(0,0, alpha=0, color=f"{v}")])
                                        for k, v in color_map.items()]
        else:
            items = [(f"{list(names)[i]}", [p.circle(0,0, alpha=0, color=f"{kv[1]}")])
                                        for i, kv in enumerate(color_map.items())
                                        if i < len(seeds)
                                        ]
        legend = bokeh.models.Legend(items=items, location="center")
        p.add_layout(legend, 'right')

        # only plotting nodes
        p.circle(source=df_, x='x',y='y', color='color', size='size',
                 line_color="white", line_width=0.1)
        p.toolbar_location="below"
        return style(p)
    return pn.Column(range_slider, pn.Row(cycle_selector, i_slider), plotter)

def ecdf_rank(opponents, my_rank,
              palette=['#cb4f70', '#1b718c', '#779e1a', '#f49044',
                      '#46308d', '#2ea58e', '#e97d86', '#7e92bd',
                      '#c21019', '#f56327', '#fab01d', '#026f69',
                      '#750a2b', '#1b9e77', '#d95f02', '#7570b3',
                      '#e7298a', '#66a61e'],
              x_range=None, title=None
              ):
    """
    Plot ECDF of everyone's nodal choices.
    Note that this rank might not reflect their method of selection,
    this is mainly for diagnostic purposes.

    Arguments
    -------------------------------------
    opponents : dictionary loaded in from json file {team name : list of list of seeds}
    my_rank : dictionary of {node : rank}, node in graph, and integer rank

    Returns
    -------------------------------------
    ECDF of selected nodes with control of iterations over 50.
    """
    iteration_slider = pn.widgets.IntSlider(start=1, end=50, value=1, name="iteration",width=785)
    @pn.depends(iteration_slider.param.value)
    def ecdf_plotter(iteration=1):
        p = bokeh.plotting.figure(height=400, width=1000, title=title, x_range=x_range)
        for a, _ in enumerate(opponents.items()):
            k, v = _
            iqplot.ecdf(data=np.array([my_rank[i] for i in v[iteration-1]]),
                        p=p, style='staircase', palette=[palette[a]],legend_label=k)

            others = set()
            for _k, _v in opponents.items():
                if _k != k:
                    _data = np.array([my_rank[i] for i in _v[iteration-1]])
                    for val in _data:
                        others.add(val)
            data = np.array([my_rank[i] for i in v[iteration-1]])
            xs, ys = np.sort(data), np.arange(1, len(data)+1) / len(data)
            d_plot = dict(zip(xs, ys))
            keep_xs = [val for val in xs if val not in others]
            keep_ys = [d_plot[x] for x in keep_xs]
            p.circle(x=keep_xs, y=keep_ys,
                fill_alpha=0.0, color=palette[a], size=8,
                line_width=2.0, legend_label=k)
        return p
    return pn.Column(iteration_slider, ecdf_plotter)



import setuptools

setuptools.setup(
    name='pandemaniac',
    version='1.0.3',
    author='Rosita Fu',
    author_email='rosita.fu99@gmail.com',
    url="https://github.com/atisor73/pandemaniac",
    description='CS144 Pandemaniac iterative visualizer',
    long_description="Pandemaniac is a package for visualizing the history of competing seeds on a multiplayer graph for the ‘pandemaniac’ project in CS144 at Caltech.",
    packages=setuptools.find_packages(),
    install_requires=["numpy", "pandas", "bokeh>=2.4.2",
                      "panel>=0.12.6", "networkx", "iqplot"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
